#! /usr/bin/python

# -*- coding: utf-8 -*-
"""
Created on Fri Sep  6 21:48:35 2013

@author: edwin
"""
import csv, sqlite3
import numpy as np
import matplotlib.pyplot as plt

class MyStdv:
    def __init__(self):
        self.count = 0
        self.s = 0.0
        self.s2 = 0.0
        
    def step(self, x):
        self.count += 1
        self.s += x
        self.s2 += x*x
        
    def finalize(self):
        return np.sqrt((self.s2 / self.count) - (self.s / self.count) ** 2)

con = sqlite3.connect(":memory:")
con.create_aggregate("stdv", 1, MyStdv)

cur = con.cursor()
cur.execute("CREATE TABLE t (algo TEXT, image TEXT, nbits INTEGER, nthread INTEGER, size INTEGER, time REAL);")
cur.execute("CREATE TABLE algos (serial TEXT, parallel TEXT);") # serial <=> parallel

with open('log.txt','rb') as fin:
    # csv.DictReader uses first line in file for column headings by default
    dr = csv.reader(fin)
    dr.next() # skip first line
    to_db = [i for i in dr]

cur.executemany("INSERT INTO t (algo, image, nbits, nthread, size, time) VALUES (?, ?, ?, ?, ?, ?);", to_db)
cur.executemany("INSERT INTO algos (serial, parallel) VALUES (?, ?);",
                [("maxtree_serial_hqueue", "maxtree_parallel_hqueue"),
                 ("maxtree_serial_nister", "maxtree_parallel_pqueue"),
                 ("maxtree_parallel_ufind_line", "maxtree_parallel_ufind_line"),
                 ("maxtree_serial_ufindrank", "maxtree_parallel_ufindrank"),
                 ("maxtree_serial_ufind_wlc", "maxtree_parallel_ufind")])
                 


con.commit()

column2labels = {
    "size" : "Number of pixels $(\\times 10^6)$",
    "mean" : "Time (ms)",
    "speedup" : "Speed Up",
    "nbits" : "Number of bits",
    "nthread" : "Number of threads"
}


algos = {
    "maxtree_serial_berger" : ("Berger et al.", "b:x"),
    "maxtree_serial_nister" : ("Nister et al.", "k--o"),
    "maxtree_serial_ufindrank" : ("Berger + rank", "g-.+"),
    "maxtree_serial_hqueue" : ("Salembier et al.", "r-s"),
    "maxtree_serial_ufind_wlc" : ("Berger + level compression", "y-d"),
    "maxtree_serial_najman" : ("Najman, Couprie", "m:+"),
    "maxtree_serial_wilkinson" : ("Wilkinson", "c-+"),
    "maxtree_parallel_hqueue" : ("Salembier et al.", "r-s"),
    "maxtree_parallel_pqueue" : ("Salembier unrecursive", "c-.+"),
    "maxtree_parallel_ufind_line": ("Matas et al.", "k:*"),
    "maxtree_parallel_ufindrank": ("Berger + rank", "g-.+"),
    "maxtree_parallel_ufind" : ("Berger + level compression", "y-d")
}

def plotdata(cur):
    data = {} # algo -> ([x], [y], [err])        
    for row in cur:
        if not data.has_key(row[0]):
            data[row[0]] = ([],[],[])
        d = data[row[0]]
        d[0].append(row[1])
        d[1].append(row[2])
        d[2].append(row[3])
    plt.clf()
    for k, d in data.iteritems():
        lbl, mark = algos[k]
        plt.plot(d[0], d[1], mark, label=lbl)
        plt.errorbar(d[0], d[1], d[2], fmt=mark[0] + mark[-1])
    plt.xlabel(column2labels[cur.description[1][0]])
    plt.ylabel(column2labels[cur.description[2][0]])
    

from matplotlib.ticker import FormatStrFormatter



    
######################################
# Display by Seq Algorithm by size   #
######################################
rq = '''
SELECT algo, size / 1000000 as size, AVG(time*1000) as mean, stdv(time*1000) / 2 as std
FROM t
WHERE nthread = -1 AND nbits = 8
GROUP BY algo, size;
'''

'''
figure()
plotdata(cur.execute(rq))
pl = gca()
pl.set_xscale("log", basex=2)
pl.set_yscale("log", basey=2)
pl.set_xticks(2**arange(10))
pl.set_yticks(  50*2**arange(14) )
pl.yaxis.set_major_formatter(FormatStrFormatter('%d'))
pl.xaxis.set_major_formatter(FormatStrFormatter('%d'))
pl.set_xlim(1,256)
l = legend(loc=2, prop = {"size" : "medium"})
l.get_frame().set_visible(False)
draw()
'''

######################################
# Display by Seq Algorithm by nbits  #
######################################
rq = '''
SELECT algo, nbits, AVG(time*1000) as mean, stdv(time*1000) / 2 as std
FROM t
WHERE nthread = -1 AND size = 8388608
GROUP BY algo, nbits;
'''

'''
figure()
plotdata(cur.execute(rq))
pl = gca()
pl.set_xlim(8,32)
pl.set_xticks(range(8,34,2))
pl.set_ylim(0,3500)
pl.set_yticks(range(0,3501,500) )
l = legend(loc=4, prop = {"size" : "medium"})
l.get_frame().set_visible(False)
draw()
'''

######################################
# Display by Number of threads  #
######################################
rq = '''
SELECT algo, nthread, AVG(time*1000) as mean, stdv(time*1000) / 2 as std
FROM t
WHERE nthread != -1 AND size = 8388608 AND nbits = 8
GROUP BY algo, nthread;
'''

'''
figure()
plotdata(cur.execute(rq))
pl = gca()
pl.set_xlim(0.9,16)
pl.set_xticks(range(2,17,2))
pl.set_ylim(100,750)
l = legend(loc=1, prop = {"size" : "medium"})
l.get_frame().set_visible(False)
draw()
'''

###########################################
# Display by Number of threads (Speed up) #
###########################################
rq = '''
SELECT t2.algo, t2.nthread, avg(t1.time / t2.time) as speedup, stdv(t1.time / t2.time) / 2 as std
FROM t as t1
INNER JOIN algos ON t1.algo = algos.serial
INNER JOIN t as t2 ON t2.algo = algos.parallel AND t1.image = t2.image AND t1.size = t2.size AND t1.nbits = t2.nbits
WHERE t1.nthread <= 1 and t1.nbits = 8 AND t1.size = 8388608
GROUP BY t2.algo, t2.nthread;
'''

'''
figure()
plotdata(cur.execute(rq))
pl = gca()
pl.set_xlim(0.95,16)
pl.set_xticks(range(2,17,2))
pl.set_ylim(0.5,4)
l = legend(loc=4, prop = {"size" : "medium"})
l.get_frame().set_visible(False)
draw()
'''

######################################
# Display Parallel by Nbits  #
######################################
rq = '''
SELECT algo, nbits, AVG(time*1000) as mean, stdv(time*1000) / 2 as std
FROM t
WHERE nthread != -1 AND size = 8388608 AND nthread = 8
GROUP BY algo, nbits;
'''

'''
figure()
plotdata(cur.execute(rq))
pl = gca()
pl.set_xlim(8,18)
pl.set_ylim(0,2500)
l = legend(loc=1, prop = {"size" : "medium"})
l.get_frame().set_visible(False)
draw()
'''
### END

algo_arg_names = {
    "salembier" : ["maxtree_serial_hqueue", "maxtree_parallel_hqueue"],                                                                                                                                                                
    "berger" : ["maxtree_serial_berger"],                                                                                                                                                                        
    "berger_rank" : ["maxtree_serial_ufindrank", "maxtree_parallel_ufindrank"],                                                                                                                                                                       
    "berger_lc" : ["maxtree_serial_ufind_wlc"],                                                                                                                                                              
    "nister" : ["maxtree_serial_nister"],                                                                                                                                                                    
    "wilkinson" : ["maxtree_serial_wilkinson"],                                                                                                                                                                                 
    "najman" : ["maxtree_serial_najman"],                                                                                                                                                                           
    "salembier_unrec" : ["maxtree_parallel_pqueue"],                                                                                                                                                               
    "matas" : ["maxtree_parallel_ufind_line"]
   }



import argparse
parser = argparse.ArgumentParser(description="Demo program for maxtree")
parser.add_argument("plottype", default="time", choices=("time","speedup"))
parser.add_argument("--by", required=True, choices=("nthread", "nbits", "size"))
parser.add_argument("--algorithms", nargs="+", help="Algorithms to plot")
parser.add_argument("--images", nargs="+", help="Filter with an image set used for benching")
parser.add_argument("--nthread", nargs="+", type=int, help="Filter with the number of threads (-1 for sequential algorithms)")
parser.add_argument("--nbits", nargs="+", type=int, help="Filter with this number of bits")
parser.add_argument("--size", nargs="+", type=int, help="Filter with this size (in Mb)")
parser.add_argument('-o', required=True, help="Output filename")

#FIXME: check arguments !

args = parser.parse_args()
rq  = "SELECT algo, %s, AVG(time * 1000) as mean, stdv(time*1000) / 2 as std \n" % args.by
rq += "FROM t \n"
rq += "WHERE (1) "

SQLTuple = lambda x: str(tuple(x) + ("NULL",) )

if args.algorithms:
    algs = sum(map(algo_arg_names.get, args.algorithms), [])
    rq += "AND algo IN %s " % SQLTuple(algs)
if args.images:
    rq += "AND image IN %s " % SQLTuple(args.images)
if args.nthread:
    rq += "AND nthread IN %s " % SQLTuple(args.nthread)
if args.nbits:
    rq += "AND nbits IN %s " % SQLTuple(args.nbits)
if args.size:
    rq += "AND size IN %s " % SQLTuple( [x * 1024 * 1024 for x in args.size] )
    
rq += "\n"
rq += "GROUP BY algo, %s --" % args.by
    
print rq

plotdata(cur.execute(rq))
l = plt.legend(loc=0, prop = {"size" : "medium"})
l.get_frame().set_visible(False)

# Un peu de formatage
pl = plt.gca()
if  args.by == "size":
    pl.set_xscale("log", basex=2)
    pl.set_yscale("log", basey=2)

plt.savefig(args.o)


















