#ifndef MLN_MORPHO_GRADIENT_HPP
# define MLN_MORPHO_GRADIENT_HPP

# include <mln/core/images/image.hpp>
# include <mln/accu/accumulators/minmax.hpp>


namespace mln
{

  namespace morphp
  {


    template <class I, class N, class Compare = std::less<mln_value(I)> >
    mln_concrete(I)
    gradient(const Image<I>& input,
	     const Neighborhood<N>& nbh,
	     const Compare& cmp = Compare());


    template <class I, class N, class O, class Compare = std::less<mln_value(I)> >
    O&
    gradient(const Image<I>& input,
	     const Neighborhood<N>& nbh,
	     Image<O>& out,
	     const Compare& cmp = Compare());


    /*************************/
    /***  Implementation   ***/
    /*************************/

    namespace impl
    {

      template <class I, class N, class O, class Compare>
      void
      gradient(const I& input,
	       const N& nbh,
	       O& out,
	       const Compare& cmp)
      {
	typedef accu::accumulator::minmax<mln_value(I), Compare> A;
	A acc(cmp);
	accfpfn(input, nbh, out, true, acc,
		[] (const A& acc) {
		  return extract(accu, mln::accu::features::max ()) -
		    extract(accu, mln::accu::features::min ());
		});
      }


    }


    template <class I, class N, class O, class Compare>
    O&
    gradient(const Image<I>& input,
	     const Neighborhood<N>& nbh,
	     Image<O>& out,
	     const Compare& cmp)
    {
      impl::gradient(exact(input), exact(nbh), exact(out), cmp);
      return exact(out);
    }

    template <class I, class N, class O, class Compare>
    mln_concrete(I)
    gradient(const Image<I>& input,
	     const Neighborhood<N>& nbh,
	     const Compare& cmp)
    {
      mln_concrete(I) out;
      resize(out, input);
      impl::gradient(exact(input), exact(nbh), out, cmp);
      return exact(out);
    }

  }

}

#endif // !MLN_MORPHO_GRADIENT_HPP
